/*
 * Copyright (c) 2010 Fabio Checconi, Luigi Rizzo, Paolo Valente
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifdef _KERNEL
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/module.h>
#include <net/if.h>	/* IFNAMSIZ */
#include <netinet/in.h>
#include <netinet/ip_var.h>		/* ipfw_rule_ref */
#include <netinet/ip_fw.h>	/* flow_id */
#include <netinet/ip_dummynet.h>
#include <netinet/ipfw/dn_heap.h>
#include <netinet/ipfw/ip_dn_private.h>
#include <netinet/ipfw/dn_sched.h>
#else
#include <dn_test.h>
#endif

#ifdef QFQP_DEBUG
struct qfqp_sched;
static void dump_sched(struct qfqp_sched *q, const char *msg);
#define	NO(x)	x
#else
#define NO(x)
#endif
#define DN_SCHED_QFQP	5 // XXX Where?
typedef uint64_t	u64;
typedef	unsigned int	bitmap;

#ifndef __FreeBSD__
static inline unsigned long __fls(unsigned long word)
{
	asm("bsr %1,%0"
	    : "=r" (word)
	    : "rm" (word));
	return word;
}
#else
static inline unsigned long __fls(unsigned long word)
{
	return fls(word) - 1;
}
#endif

#ifdef QFQP_DEBUG
int test_bit(int ix, bitmap *p)
{
	NO(if (ix < 0 || ix > 31)
		   D("bad index %d", ix));
	return *p & (1<<ix);
}
void __set_bit(int ix, bitmap *p)
{
	NO(if (ix < 0 || ix > 31)
		   D("bad index %d", ix));
	*p |= (1<<ix);
}
void __clear_bit(int ix, bitmap *p)
{
	NO(if (ix < 0 || ix > 31)
		   D("bad index %d", ix));
	*p &= ~(1<<ix);
}
#else /* !QFQP_DEBUG */
/* XXX do we have fast version, or leave it to the compiler ? */
#define test_bit(ix, pData)	((*pData) & (1<<(ix)))
#define __set_bit(ix, pData)	(*pData) |= (1<<(ix))
#define __clear_bit(ix, pData)	(*pData) &= ~(1<<(ix))
#endif /* !QFQP_DEBUG */
/*-------------------------------------------*/
/*

Virtual time computations.

S, F and V are all computed in fixed point arithmetic with
FRAC_BITS decimal bits.

   QFQP_MAX_INDEX is the maximum index allowed for a group. We need
  	one bit per index.
   QFQP_MAX_WSHIFT is the maximum power of two supported as a weight.
   The layout of the bits is as below:
  
                   [ MTU_SHIFT ][      FRAC_BITS    ]
                   [ MAX_INDEX    ][ MIN_SLOT_SHIFT ]
  				 ^.__grp->index = 0
  				 *.__grp->slot_shift
  
   where MIN_SLOT_SHIFT is derived by difference from the others.

The max group index corresponds to Lmax/w_min, where
Lmax=M * (1<<MTU_SHIFT), w_min = 1 .
From this, and knowing how many groups (MAX_INDEX) we want,
we can derive the shift corresponding to each group.

Because we often need to compute
	F = S + len/w_i  and V = V + len/wsum
instead of storing w_i store the value
	inv_w = (1<<FRAC_BITS)/w_i
so we can do F = S + len * inv_w * wsum.
We use W_TOT in the formulas so we can easily move between
static and adaptive weight sum.

The per-scheduler-instance data contain all the data structures
for the scheduler: bitmaps and bucket lists.

 */
/*
 * Maximum number of consecutive slots occupied by backlogged aggregates
 * inside a group. This is approx lmax/lmin + 5.
 * XXX check because it poses constraints on MAX_INDEX
 */
#define QFQP_MAX_SLOTS	32
/*
 * Shifts used for class<->group mapping. Class weights are
 * in the range [1, QFQP_MAX_WEIGHT], we to map each class i to the
 * group with the smallest index that can support the L_i / r_i
 * configured for the class.
 *
 * grp->index is the index of the group; and grp->slot_shift
 * is the shift for the corresponding (scaled) sigma_i.
 *
 * When computing the group index, we do (len<<FP_SHIFT)/weight,
 * then compute an FLS (which is like a log2()), and if the result
 * is below the MAX_INDEX region we use 0 (which is the same as
 * using a larger len).
 */
#define QFQP_MAX_INDEX		19
#define QFQP_MAX_WSHIFT		16	/* log2(max_weight) */

#define	QFQP_MAX_WEIGHT		(1<<QFQP_MAX_WSHIFT)
#define QFQP_MAX_WSUM		(2*QFQP_MAX_WEIGHT)

#define FRAC_BITS		30	/* fixed point arithmetic */
#define ONE_FP			(1UL << FRAC_BITS)

#define MAX_CLASSES_PER_AGG_SHIFT       3
#define MAX_CLASSES_PER_AGG             (1<<MAX_CLASSES_PER_AGG_SHIFT)

#define QFQP_MTU_SHIFT		11	/* log2(max_len) */
/* log2(max_len * max_classes_per_agg) */
#define QFQP_MAXB_SHIFT	        (QFQP_MTU_SHIFT + MAX_CLASSES_PER_AGG_SHIFT)
#define QFQP_MIN_SLOT_SHIFT	(FRAC_BITS + QFQP_MAXB_SHIFT - QFQP_MAX_INDEX)

//#define TAIL_INSERT
//#define HEAVYVERSION

/*
 * Possible group states, also indexes for the bitmaps array in
 * struct qfqp_queue. We rely on ER, IR, EB, IB being numbered 0..3
 */
enum qfqp_state { ER, IR, EB, IB, QFQP_MAX_STATE };

struct qfqp_class {
	struct dn_queue q;		/* Standard queue */
	int credit;			/* Number of bytes to transmit */
	struct qfqp_class *clnext;	/* List of backlogged classes */
	struct qfqp_aggregate *parent;  /* Aggregate we belong to */
};

struct qfqp_group;
/*
 * additional queue info. Some of this info should come from
 * the flowset, we copy them here for faster processing.
 * This is an overlay of the struct dn_queue
 */
struct qfqp_aggregate {
	uint64_t S, F;		/* aggregate timestamps (exact) */
	struct qfqp_aggregate *next; /* Link for the slot list. */

	/* group we belong to. In principle we would need the index,
	 * which is log_2(lmax/weight), but we never reference it
	 * directly, only the group.
	 */
	struct qfqp_group *grp;

	/* these are copied from flowset. */
	uint32_t	class_weight; /* weight of each class in this aggr. */
	int 	lmax;  /* Max pkt size for the classes in this
				  aggr., used also as quantum */

	/* these are computed from the above parameters. */
	uint32_t	inv_w; /* ONE_FP/(sum of weights of classes in aggr.) */
	uint32_t 	budgetmax;  /* Max budget for this aggregate */
	uint32_t        budget;     /* Current budget */

	uint32_t num_classes; /* Number of classes in this aggregate */
	struct qfqp_class *rr_head, *rr_tail;  /* Pointers to drr queue */

	struct qfqp_aggregate *nonfull_next; /* Link for the non-full list. */
};

/* Group descriptor, see the paper for details.
 * Basically this contains the bucket lists
 */
struct qfqp_group {
	uint64_t S, F;			/* group timestamps (approx). */
	unsigned int slot_shift;	/* Slot shift. */
	unsigned int index;		/* Group index. */
	unsigned int front;		/* Index of the front slot. */

	/*
	 * Bitmap with non-empty slots.  Using unsigned long as most
	 * bitops in Linux use it (i'd rather use u32 or u64 XXX)
	 */
	bitmap full_slots;

	/* Array of lists of active aggregates. */
	struct qfqp_aggregate *slots[QFQP_MAX_SLOTS];
#ifdef TAIL_INSERT
	/* Array of tails of lists of active aggregates. */
	struct qfqp_aggregate *slot_tails[QFQP_MAX_SLOTS];
#endif
};

/* scheduler instance descriptor. */
struct qfqp_sched {
	uint64_t	oldV, V;		/* Precise virtual times. */
	struct qfqp_aggregate  *active_agg;    /* Aggregate under service */
	uint32_t        num_queued_agg; /* Number of aggregates queued */
	uint32_t	wsum;		/* weight sum */
	uint32_t	i_wsum;		/* ONE_FP/w_sum */
	NO(uint32_t	queued;	/* debugging */
	uint32_t	loops;	/* debugging */)
	bitmap bitmaps[QFQP_MAX_STATE];	/* Group bitmaps. */
	struct qfqp_group groups[QFQP_MAX_INDEX + 1]; /* The groups. */
	struct qfqp_aggregate *nonfull_agg_list; /* List of aggregates that
					    have still room for other
					    classes */

#if 0
	NO("int num_unblock_groups, num_update_start, num_make_eligible,"
	   "num_update_eligible, num_rotate_queue, num_rr_dequeue,"
	   "num_choose_aggregate, num_enqueue_aggregate, num_dequeue,"
	   "num_enqueue");
#endif
};

/*---- support functions ----------------------------*/

/* Generic comparison function, handling wraparound. */
static inline int qfqp_gt(u64 a, u64 b)
{
	return (int64_t)(a - b) > 0;
}

/* Round a precise timestamp to its slotted value. */
static inline u64 qfqp_round_down(u64 ts, unsigned int shift)
{
	return ts & ~((1ULL << shift) - 1);
}

/* return the pointer to the group with lowest index in the bitmap */
static inline struct qfqp_group *qfqp_ffs(struct qfqp_sched *q,
					unsigned long bitmap)
{
	int index = ffs(bitmap) - 1; // zero-based
	ND("index %d", index);
	return &q->groups[index];
}

/*
 * Calculate an aggregate index, given its weight and maximum budget.
 * index = log_2(maxlen/weight) but we need to apply the scaling.
 * This is used only once at flow creation.
 */
static int qfqp_calc_index(uint32_t inv_w, unsigned int maxlen)
{
	u64 slot_size = (u64)maxlen * inv_w;
	unsigned long size_map;
	int index = 0;

	size_map = (unsigned long)(slot_size >> QFQP_MIN_SLOT_SHIFT);
	if (!size_map)
		goto out;

	index = __fls(size_map) + 1;	// basically a log_2()
	index -= !(slot_size - (1ULL << (index + QFQP_MIN_SLOT_SHIFT - 1)));

	if (index < 0)
		index = 0;

out:
	ND("W = %d, L = %d, I = %d\n", ONE_FP/inv_w, maxlen, index);
	return index;
}

static inline struct qfqp_aggregate *
create_aggregate(struct qfqp_sched *q,
		      uint32_t weight, uint32_t lmax)
{
	struct qfqp_aggregate *agg = calloc(1, sizeof(struct qfqp_aggregate));
	
	agg->class_weight = weight;
	agg->lmax = lmax;
	if (agg->class_weight == 0 || agg->class_weight > QFQP_MAX_WEIGHT) {
		ND("rounding weight to 1");
		agg->class_weight = 1;
	}

	/* agg is empty, insert into the list of non-full aggregates */
	agg->nonfull_next = q->nonfull_agg_list ;
	q->nonfull_agg_list = agg ;

	return agg;
}

static inline struct qfqp_aggregate *
find_create_aggregate(struct dn_queue *_q)
{
	struct qfqp_sched *q = (struct qfqp_sched *)(_q->_si + 1);
	struct qfqp_aggregate *agg ;
	/* import parameters from the flowset. They should be correct
	 * already.
	 */

	uint32_t weight = _q->fs->fs.par[0];
	uint32_t lmax = _q->fs->fs.par[1];

	if (lmax < sizeof(struct mbuf)) {
		D("lmax too small");
		exit(1);
	}
	/* look for a matching aggregate in the list of non-full aggregates */
	for (agg = q->nonfull_agg_list ;
	     agg != NULL && (agg->class_weight != weight || agg->lmax != lmax) ;
	     agg = agg->nonfull_next)
		;

	if (agg != NULL) // aggregate found
		return agg ; // return it
	else /* invoke create_aggregate to create a new one and return it */
		return create_aggregate(q, weight, lmax) ;
}

static inline void
remove_from_list(struct qfqp_aggregate **nonfull_list,
		 struct qfqp_aggregate *agg)
{
	struct qfqp_aggregate *prev_agg ;
	
	ND("nonfull_list %p agg %p", nonfull_list, agg);
	if (agg == *nonfull_list) {
		*nonfull_list = agg->nonfull_next;
		ND("head-remove");
		return;
	}

	/* look for a matching aggregate in the list of non-full aggregates */
	for (prev_agg = *nonfull_list ;
	     prev_agg != NULL && prev_agg->nonfull_next != agg;
	     prev_agg = prev_agg->nonfull_next)
		;

	if (prev_agg == NULL) { // previous aggregate not found
		D("prev_agg not found");
		exit(1);
	}
	prev_agg->nonfull_next = agg->nonfull_next;
}

/* Append a queue to the rr list */
static inline void
rr_append(struct qfqp_class *cl, struct qfqp_aggregate *agg)
{
	/* append to the tail */
	if (agg->rr_head == NULL)
		agg->rr_head = cl;
	else
		agg->rr_tail->clnext = cl;
	agg->rr_tail = cl;		/* advance the tail pointer */
	cl->clnext = agg->rr_head;	/* make it circular */
}

/* Remove the head queue from circular list. */
static inline void
rr_remove_head(struct qfqp_aggregate *agg)
{
#ifndef HEAVYVERSION
	if (agg->rr_head == agg->rr_tail) {
		agg->rr_head = agg->rr_tail = NULL;
		return;
	}

	agg->rr_head = agg->rr_head->clnext;
	agg->rr_tail->clnext = agg->rr_head;
#else
	struct qfqp_class *old_head = agg->rr_head;

	if (agg->rr_head == agg->rr_tail) {
		agg->rr_head->clnext = NULL;
		agg->rr_head = agg->rr_tail = NULL;
		return;
	}

	agg->rr_head = agg->rr_head->clnext;
	agg->rr_tail->clnext = agg->rr_head;
	old_head->clnext = NULL;
#endif
}

/*---- end support functions ----*/

/*-------- API calls --------------------------------*/
/*
 * Validate and copy parameters from flowset.
 */
static int
qfqp_new_queue(struct dn_queue *_q)
{
	struct qfqp_sched *q = (struct qfqp_sched *)(_q->_si + 1);
	struct qfqp_class *cl = (struct qfqp_class *)_q;
	struct qfqp_aggregate *agg = find_create_aggregate(_q);
	uint32_t agg_weight;
	int i;

	if (q->wsum + agg->class_weight > QFQP_MAX_WSUM) {
		D("max wsum exceeded");
		exit(1);
	}

	agg->num_classes++ ;
	if (agg->num_classes == MAX_CLASSES_PER_AGG)
		remove_from_list(&q->nonfull_agg_list, agg);

	agg_weight = agg->num_classes * agg->class_weight;
	agg->inv_w = ONE_FP/agg_weight;

	ND("weight %d", agg->class_weight);

	q->wsum += agg->class_weight;
	q->i_wsum = ((1UL<<FRAC_BITS)/q->wsum);

	agg->budgetmax = agg->num_classes * agg->lmax;

	i = qfqp_calc_index(agg->inv_w, agg->budgetmax);

	agg->grp = &q->groups[i];

	cl->credit = agg->lmax;
	cl->parent = agg;

	if (_q->mq.head != NULL) {
		/* Queue NOT empty, insert in the queue list */
		rr_append(cl, agg);
	}
	return 0;
}

/* remove an empty queue */
static int
qfqp_free_queue(struct dn_queue *_q)
{
	D("qfqp_free_queue not implemented") ;
	exit(1) ;

	return 0;
}

/* Calculate a mask to mimic what would be ffs_from(). */
static inline unsigned long
mask_from(unsigned long bitmap, int from)
{
	return bitmap & ~((1UL << from) - 1);
}

/*
 * The state computation relies on ER=0, IR=1, EB=2, IB=3
 * First compute eligibility comparing grp->S, q->V,
 * then check if someone is blocking us and possibly add EB
 */
static inline unsigned int
qfqp_calc_state(struct qfqp_sched *q, struct qfqp_group *grp)
{
	/* if S > V we are not eligible */
	unsigned int state = qfqp_gt(grp->S, q->V);
	unsigned long mask = mask_from(q->bitmaps[ER], grp->index);
	struct qfqp_group *next;

	if (mask) {
		next = qfqp_ffs(q, mask);
		if (qfqp_gt(grp->F, next->F))
			state |= EB;
	}

	return state;
}

/*
 * In principle
 *	q->bitmaps[dst] |= q->bitmaps[src] & mask;
 *	q->bitmaps[src] &= ~mask;
 * but we should make sure that src != dst
 */
static inline void
qfqp_move_groups(struct qfqp_sched *q, unsigned long mask, int src, int dst)
{
	q->bitmaps[dst] |= q->bitmaps[src] & mask;
	q->bitmaps[src] &= ~mask;
}

static inline void
qfqp_unblock_groups(struct qfqp_sched *q, int index, u64 old_finish)
{
	unsigned long mask = mask_from(q->bitmaps[ER], index + 1);
	struct qfqp_group *next;

	ND(q->num_unblock_groups++);

	if (mask) {
		next = qfqp_ffs(q, mask);
		ND("next->F %lu oldfinish %lu",
		  (next->F)>>QFQP_MIN_SLOT_SHIFT,
		  old_finish>>QFQP_MIN_SLOT_SHIFT);
		if (!qfqp_gt(next->F, old_finish))
			return;
	}
	mask = (1UL << (index)) - 1;
	ND("index %d mask %x qnext %p", index, mask, next);
	qfqp_move_groups(q, mask, EB, ER);
	qfqp_move_groups(q, mask, IB, IR);
}

/*
 * perhaps
 *
	oldV ^= q->V;
	oldV >>= QFQP_MIN_SLOT_SHIFT;
	if (oldV) {
		...
	}
 *
 */
static inline void
qfqp_make_eligible(struct qfqp_sched *q)
{
	unsigned long mask, vslot, old_vslot;

	ND(q->num_make_eligible++);

	vslot = q->V >> QFQP_MIN_SLOT_SHIFT;
	old_vslot = q->oldV >> QFQP_MIN_SLOT_SHIFT;

	if (vslot != old_vslot) {
		mask = (2ULL << (__fls(vslot ^ old_vslot))) - 1;
		qfqp_move_groups(q, mask, IR, ER);
		qfqp_move_groups(q, mask, IB, EB);
	}
}

/*
 * XXX we should make sure that slot becomes less than 32.
 * This is guaranteed by the input values.
 * roundedS is always agg->S rounded on grp->slot_shift bits.
 */
static inline void
qfqp_slot_insert(struct qfqp_group *grp, struct qfqp_aggregate *agg,
		 u64 roundedS)
{
	u64 slot = (roundedS - grp->S) >> grp->slot_shift;
	unsigned int i = (grp->front + slot) % QFQP_MAX_SLOTS;

	ND("agg %p S %lu roundedS-grp->S %lu budg %lu",
	  agg, agg->S>>grp->slot_shift,
	  (roundedS - grp->S)>>grp->slot_shift,
	  agg->budget);

	NO(if (slot >= QFQP_MAX_SLOTS) {
			exit(1);});

#ifndef TAIL_INSERT
	agg->next = grp->slots[i];
	grp->slots[i] = agg;
#else
	agg->next = 0;
	if (grp->slot_tails[i] != NULL)
		grp->slot_tails[i]->next = agg;
	else
		grp->slots[i] = agg;
	grp->slot_tails[i] = agg;
#endif
	__set_bit(slot, &grp->full_slots);
}

/*
 * remove the entry from the slot
 */
static inline void
qfqp_front_slot_remove(struct qfqp_group *grp)
{
	struct qfqp_aggregate **h = &grp->slots[grp->front];

	*h = (*h)->next;
	if (!*h) {
		__clear_bit(0, &grp->full_slots);
#ifdef TAIL_INSERT
		grp->slot_tails[grp->front] = NULL;
#endif
	}
}

/*
 * Returns the first non-empty queue in a group. As a side effect,
 * adjust the bucket list so the first non-empty bucket is at
 * position 0 in full_slots.
 */
static inline struct qfqp_aggregate *
qfqp_slot_scan(struct qfqp_group *grp)
{
	int i;

	ND("grp %d full %x", grp->index, grp->full_slots);
	if (!grp->full_slots)
		return NULL;

	i = ffs(grp->full_slots) - 1; // zero-based
	if (i > 0) {
		grp->front = (grp->front + i) % QFQP_MAX_SLOTS;
		grp->full_slots >>= i;
	}

	return grp->slots[grp->front];
}

/*
 * adjust the bucket list. When the start time of a group decreases,
 * we move the index down (modulo QFQP_MAX_SLOTS) so we don't need to
 * move the objects. The mask of occupied slots must be shifted
 * because we use ffs() to find the first non-empty slot.
 * This covers decreases in the group's start time, but what about
 * increases of the start time ?
 * Here too we should make sure that i is less than 32
 */
static inline void
qfqp_slot_rotate(struct qfqp_sched *q, struct qfqp_group *grp, u64 roundedS)
{
	unsigned int i = (grp->S - roundedS) >> grp->slot_shift;

	grp->full_slots <<= i;
	grp->front = (grp->front - i) % QFQP_MAX_SLOTS;
}


static inline void
qfqp_update_eligible(struct qfqp_sched *q)
{
	bitmap ineligible;

	ND(q->num_update_eligible++);

	ineligible = q->bitmaps[IR] | q->bitmaps[IB];
	if (ineligible) {
		if (!q->bitmaps[ER]) {
			struct qfqp_group *grp;
			grp = qfqp_ffs(q, ineligible);
			if (qfqp_gt(grp->S, q->V))
				q->V = grp->S;
		}
		qfqp_make_eligible(q);
	}
}

static inline void
rotate_rr_queue(struct qfqp_aggregate *agg)
{

	if (agg->rr_head == NULL)
		return; /* empty queue */

	agg->rr_head = agg->rr_head->clnext;
	agg->rr_tail = agg->rr_tail->clnext;
}

static inline struct mbuf *
rr_dequeue(struct qfqp_aggregate *agg)
{
	uint64_t len;
#ifndef HEAVYVERSION
	struct qfqp_class *cl = agg->rr_head;
	struct mbuf *m;

	m = cl->q.mq.head;
	len = m->m_pkthdr.len;

	cl->credit -= (int) len;
	dn_dequeue(&cl->q);
	ND("dn_dequeue %p credit %d %p flow_id %u length %u",
	   cl, cl->credit,
	   m, m->flow_id, cl->q.ni.length);
	
	if (cl->q.mq.head == NULL)
		rr_remove_head(agg);
	else if (cl->credit < cl->q.mq.head->m_pkthdr.len) {
		cl->credit += agg->lmax;
		rotate_rr_queue(agg);
	}

	return m;
#else
	struct qfqp_class *cl;

	while ( (cl = agg->rr_head) ) {
		struct mbuf *m = cl->q.mq.head;
		if ( m == NULL) {
			/* empty queue, remove from list */
			rr_remove_head(agg);
			continue;
		}
		len = m->m_pkthdr.len;

		if (len > cl->credit) {
			/* Packet too big */
			cl->credit += agg->lmax;
			/* Try next queue */
			rotate_rr_queue(agg);
		} else {
			cl->credit -= len;
			ND("dn_dequeue %p credit %d %p flow_id %u",
			  cl, cl->credit,
			  m, m->flow_id);
			return dn_dequeue(&cl->q);
		}
	}

	/* no packet to dequeue*/
	return NULL;
#endif

}

static inline struct qfqp_aggregate *
choose_next_aggregate(struct qfqp_sched *q)
{
	struct qfqp_group *grp;
	struct qfqp_aggregate *agg, *new_front_agg;
	u64 old_F;

	qfqp_update_eligible(q);
	q->oldV = q->V;

	if (!q->bitmaps[ER]) {
		NO(if (q->queued)
			   dump_sched(q, "start choose_next_aggregate"););
		return NULL;
	}

	grp = qfqp_ffs(q, q->bitmaps[ER]);
	old_F = grp->F;

	agg = grp->slots[grp->front];

	qfqp_front_slot_remove(grp);
	q->num_queued_agg--;

	new_front_agg = qfqp_slot_scan(grp);

	if (new_front_agg == NULL) { /* group gone, remove from ER */
		ND("removing group from ER");
		__clear_bit(grp->index, &q->bitmaps[ER]);
	} else {
		u64 roundedS =
			qfqp_round_down(new_front_agg->S,
					grp->slot_shift);
		unsigned int s;
		
		if (grp->S == roundedS)
			return agg;
		grp->S = roundedS;
		grp->F = roundedS + (2ULL << grp->slot_shift);
		/* remove from ER and put in the new set */
		__clear_bit(grp->index, &q->bitmaps[ER]);
		s = qfqp_calc_state(q, grp);
		__set_bit(grp->index, &q->bitmaps[s]);
	}
	/* we need to unblock even if the group has gone away */
	qfqp_unblock_groups(q, grp->index, old_F);
	ND("agg %p S %lu F %lu V %lu V-S %lu V-old %lu",
	  agg, agg->S>>grp->slot_shift, (agg->F)>>grp->slot_shift,
	  (q->V)>>grp->slot_shift, (q->V-agg->S)>>grp->slot_shift,
	  (q->V-q->oldV)>>grp->slot_shift);
	NO(if (!(q->bitmaps[ER] | q->bitmaps[EB] |
		 q->bitmaps[IR] | q->bitmaps[IB]) && q->num_queued_agg > 0)
		   dump_sched(q, "choose_next_aggregate after unblock");)

	return agg;
}

static inline void
charge_actual_service(struct qfqp_aggregate *agg,
		      uint32_t service_received)
{
	agg->F = agg->S + (u64)service_received * agg->inv_w;
}

/*
 * Assign a reasonable start time for a new flow k in group i.
 * Admissible values for \hat(F) are multiples of \sigma_i
 * no greater than V+\sigma_i . Larger values mean that
 * we had a wraparound so we consider the timestamp to be stale.
 *
 * If F is not stale and F >= V, then we set S = F. Otherwise we
 * should assign S = V, but this may violate the ordering in EB (see
 * the paper). To preserve the ordering, if we have groups in ER, set
 * S to the F_j of the first group j which would be blocking us.  We
 * are guaranteed not to move S backward because otherwise our group i
 * would still be blocked.
 */
static inline void
qfqp_update_start(struct qfqp_sched *q, struct qfqp_aggregate *agg)
{
	unsigned long mask;
	uint32_t limit, roundedF;
	int slot_shift = agg->grp->slot_shift;


	roundedF = qfqp_round_down(agg->F, slot_shift);
	limit = qfqp_round_down(q->V, slot_shift) + (1UL << slot_shift);

	if (!qfqp_gt(agg->F, q->V) || qfqp_gt(roundedF, limit)) {
		/* timestamp was stale */
		mask = mask_from(q->bitmaps[ER], agg->grp->index);
		if (mask) {
			struct qfqp_group *next = qfqp_ffs(q, mask);
			if (qfqp_gt(roundedF, next->F)) {
				ND(q->num_update_start++);
				if (qfqp_gt(roundedF, next->F))
					agg->S = next->F;
				else
					/* preserve timestamp correctness */
					agg->S = limit;
				return;
			}
		}
		agg->S = q->V;
	} else { /* timestamp is not stale */
		agg->S = agg->F;
	}
}

enum update_reason {enqueue, requeue};

static inline void
update_aggregate_ts(struct qfqp_sched *q,
		    struct qfqp_aggregate *agg, enum update_reason reason)
{
	if (reason != requeue)
		qfqp_update_start(q, agg); /* adjust start time */
	else
		agg->S = agg->F;

	agg->F = agg->S + (u64)agg->budgetmax * agg->inv_w;
}

static inline void
qfqp_enqueue_aggregate(struct qfqp_sched *q, struct qfqp_aggregate *agg)
{
	struct qfqp_group *grp;
	u64 roundedS;
	int s;

	q->num_queued_agg++;
	grp = agg->grp;
	roundedS = qfqp_round_down(agg->S, grp->slot_shift);

	/*
	 * insert agg in the correct bucket.
	 * If agg->S >= grp->S we don't need to adjust the
	 * bucket list and simply go to the insertion phase.
	 * Otherwise grp->S is decreasing, we must make room
	 * in the bucket list, and also recompute the group state.
	 * Finally, if there were no flows in this group and nobody
	 * was in ER make sure to adjust V.
	 */
	if (grp->full_slots) {
		if (!qfqp_gt(grp->S, agg->S))
			goto skip_update;
		/* create a slot for this agg->S */
		qfqp_slot_rotate(q, grp, roundedS);
		/* group was surely ineligible, remove */
		__clear_bit(grp->index, &q->bitmaps[IR]);
		__clear_bit(grp->index, &q->bitmaps[IB]);
	} else if (!q->bitmaps[ER] && qfqp_gt(roundedS, q->V) && q->active_agg == NULL)
		q->V = roundedS;
	ND(q->num_enqueue_aggregate++);

	grp->S = roundedS;
	grp->F = roundedS + (2ULL << grp->slot_shift); // i.e. 2\sigma_i
	s = qfqp_calc_state(q, grp);
	__set_bit(grp->index, &q->bitmaps[s]);
	ND("new state %d 0x%x", s, q->bitmaps[s]);
	ND("agg %p S %lu F %lu V %lu V-S %lu V-old %lu",
	  agg, agg->S>>grp->slot_shift, (agg->F)>>grp->slot_shift,
	  (q->V)>>grp->slot_shift, (q->V-agg->S)>>grp->slot_shift,
	  (q->V-q->oldV)>>grp->slot_shift);
skip_update:
	qfqp_slot_insert(grp, agg, roundedS);
}

static struct mbuf *
qfqp_dequeue(struct dn_sch_inst *si)
{
	struct qfqp_sched *q = (struct qfqp_sched *)(si + 1);
	struct qfqp_aggregate *active_agg = q->active_agg;
	struct mbuf *packet;
#ifndef HEAVYVERSION
	uint32_t pkt_len = active_agg->rr_head == NULL ?
		0 : active_agg->rr_head->q.mq.head->m_pkthdr.len;
#else
	uint32_t pkt_len;
	if (active_agg->rr_head != NULL){
		while (active_agg->rr_head != NULL &&
		       active_agg->rr_head->q.mq.head == NULL)
			rr_remove_head(active_agg);
	}
	pkt_len = active_agg->rr_head != NULL &&
		active_agg->rr_head->q.mq.head != NULL ?
		active_agg->rr_head->q.mq.head->m_pkthdr.len : 0;
#endif
	
	ND(q->num_dequeue++);

	NO(q->loops++;
	   if (!(q->bitmaps[ER] | q->bitmaps[EB] |
		 q->bitmaps[IR] | q->bitmaps[IB]) && q->num_queued_agg > 0)
		   dump_sched(q, "qfqp_dequeue"););
	
	ND("starting: active_agg %p rr_head %p budg %u q->num_queued_agg %u",
	  active_agg, active_agg->rr_head, active_agg->budget,
	  q->num_queued_agg);

#if 0
	if (pkt_len == 0 && q->num_queued_agg == 0) {
		D("anomalous call");
		exit(1);
	}
#endif
#define unlikely(x)     __builtin_expect((x),0)

	// if there is not enough budget to serve next class in the aggregate
	if (unlikely(pkt_len == 0 || active_agg->budget < pkt_len)) {
		
		// compute the service received by the aggregate
		uint32_t service_received =
			active_agg->budgetmax - active_agg->budget;

		charge_actual_service(active_agg, service_received);
		
		// recharge the budget of the aggregate
		active_agg->budget = active_agg->budgetmax;

#ifndef HEAVYVERSION
		if (active_agg->rr_head != NULL) {
#else
		if (active_agg->rr_head != NULL && 
		    active_agg->rr_head->q.mq.head != NULL) {
#endif
			// update aggregate ts to correctly requeue it
			update_aggregate_ts(q, active_agg, requeue);
			// enqueue aggregate into the aggregate scheduler
			qfqp_enqueue_aggregate(q, active_agg);
		} else 	if (!q->bitmaps[ER]) {
			NO(if (q->queued)
				   dump_sched(q, "change next agg"););
		}

			
		// if we get here, there are other aggregates queued:
                // choose the new active aggregate
		active_agg = q->active_agg = choose_next_aggregate(q);
		ND("update_agg service_rec %u new active %p",
		  service_received, active_agg);
	} // otherwise leave active aggregate unchanged
	packet = rr_dequeue(active_agg);

	NO(q->queued--);
	pkt_len = packet->m_pkthdr.len;
	active_agg->budget -= pkt_len;
	q->V += (u64)pkt_len * q->i_wsum;
	ND("agg %p S %lu F %lu V %lu V-S %lu V-old %lu budg %u",
	  active_agg, active_agg->S>>active_agg->grp->slot_shift,
	  (active_agg->F)>>active_agg->grp->slot_shift,
	  (q->V)>>active_agg->grp->slot_shift,
	  (q->V-active_agg->S)>>active_agg->grp->slot_shift,
	  (q->V-q->oldV)>>active_agg->grp->slot_shift,
	  active_agg->budget);

	ND("flow_id %d agg %p S %lu F %lu V %lu V-S %lu V-old %lu budg %u",
	  packet->flow_id, active_agg, active_agg->S>>QFQP_MIN_SLOT_SHIFT,
	  (active_agg->F)>>QFQP_MIN_SLOT_SHIFT,
	  (q->V)>>QFQP_MIN_SLOT_SHIFT,
	  (q->V-active_agg->S)>>QFQP_MIN_SLOT_SHIFT,
	  (q->V-q->oldV)>>QFQP_MIN_SLOT_SHIFT,
	  active_agg->budget);
	
	ND(if (q->num_dequeue == 4000000) {
		D("num_unblock_groups %d, num_update_start %d\n"
		  "num_make_eligible %d, num_update_eligible %d\n"
		  "num_rotate_queue %d, num_rr_dequeue %d\n"
		  "num_choose_aggregate %d, num_enqueue_aggregate %d\n"
		  "num_dequeue %d, num_enqueue %d\n",
		  q->num_unblock_groups, q->num_update_start, q->num_make_eligible,
		  q->num_update_eligible, q->num_rotate_queue, q->num_rr_dequeue,
		  q->num_choose_aggregate, q->num_enqueue_aggregate, q->num_dequeue,
		  q->num_enqueue);
		});

	return packet;
}

static int
qfqp_enqueue(struct dn_sch_inst *si, struct dn_queue *_q, struct mbuf *m)
{
	struct qfqp_sched *q = (struct qfqp_sched *)(si + 1);
	struct qfqp_class *cl = (struct qfqp_class *)_q;
	struct qfqp_aggregate *agg = cl->parent;


	NO(q->loops++;)
	/* XXX verify that the packet obeys the parameters */
	if (m != _q->mq.head) {
		if (dn_enqueue(_q, m, 0)) /* packet was dropped */
			return 1;
		NO(q->queued++);
		ND("flow_id %d", m->flow_id);

		if (m != _q->mq.head)
			return 0;
		else if (cl == agg->rr_head && m->m_pkthdr.len > cl->credit) {
			cl->credit += agg->lmax;
			rotate_rr_queue(agg);
			
			return 0;
		}
	}
	
	ND("len %d flow %p inv_w 0x%x grp %d", m->m_pkthdr.len,
		_q, agg->inv_w, agg->grp->index);

#ifdef HEAVYVERSION
	if (cl->clnext != NULL)
		return 0;
#endif

	// init credit and append class queue to the rr queue of the aggregate
	cl->credit = agg->lmax;
	rr_append(cl, agg);
	
	if (agg->rr_head != cl || q->active_agg == agg) // aggregate was not empty
		return 0; // nothing else to do
	
	ND(q->num_enqueue++);
	// if we get here, the aggregate was empty and hence was not queued in
	// the aggregate scheduler
	// recharge budget
	agg->budget = agg->budgetmax;
	
	update_aggregate_ts(q, agg, enqueue);
	if (q->active_agg == NULL)
		q->active_agg = agg;
	else if (agg != q->active_agg)
		qfqp_enqueue_aggregate(q, agg);

	return 0;
}


static int
qfqp_new_fsk(struct dn_fsk *f)
{
	ipdn_bound_var(&f->fs.par[0], 1, 1, QFQP_MAX_WEIGHT, "qfqp weight");
	ipdn_bound_var(&f->fs.par[1], 1500, 1, 2000, "qfqp maxlen");
	DX(3, "weight %d len %d\n", f->fs.par[0], f->fs.par[1]);
	return 0;
}

/*
 * initialize a new scheduler instance
 */
static int
qfqp_new_sched(struct dn_sch_inst *si)
{
	struct qfqp_sched *q = (struct qfqp_sched *)(si + 1);
	struct qfqp_group *grp;
	int i;

	for (i = 0; i <= QFQP_MAX_INDEX; i++) {
		grp = &q->groups[i];
		grp->index = i;
		grp->slot_shift = QFQP_MIN_SLOT_SHIFT + i;
	}
	return 0;
}

/*
 * QFQP scheduler descriptor
 */
static struct dn_alg qfqp_desc = {
	.type = DN_SCHED_QFQP,
	.name = "QFQP",
	.flags = DN_MULTIQUEUE,

	.si_datalen = sizeof(struct qfqp_sched),
	.q_datalen = sizeof(struct qfqp_class) - sizeof(struct dn_queue),

	.enqueue = qfqp_enqueue,
	.dequeue = qfqp_dequeue,

	.new_sched = qfqp_new_sched,
	.new_fsk = qfqp_new_fsk,

	.new_queue = qfqp_new_queue,
	.free_queue = qfqp_free_queue,
};

DECLARE_DNSCHED_MODULE(dn_qfqp, &qfqp_desc);

#ifdef QFQP_DEBUG
static void
dump_groups(struct qfqp_sched *q, uint32_t mask)
{
	int i, j;

	for (i = 0; i < QFQP_MAX_INDEX + 1; i++) {
		struct qfqp_group *g = &q->groups[i];

		if (0 == (mask & (1<<i)))
			continue;
		for (j = 0; j < QFQP_MAX_SLOTS; j++) {
			if (g->slots[j])
				D("    bucket %d %p", j, g->slots[j]);
		}
		D("full_slots 0x%x", g->full_slots);
		D("        %2d S 0x%20lu F 0x%lu %c", i,
			g->S, g->F,
			mask & (1<<i) ? '1' : '0');
	}
}

static void
dump_sched(struct qfqp_sched *q, const char *msg)
{
	D("--- in %s: ---", msg);
	ND("loops %d queued %d V 0x%llx", q->loops, q->queued, q->V);
	D("    ER 0x%08x", q->bitmaps[ER]);
	D("    EB 0x%08x", q->bitmaps[EB]);
	D("    IR 0x%08x", q->bitmaps[IR]);
	D("    IB 0x%08x", q->bitmaps[IB]);
	//dump_groups(q, 0xffffffff);
};
#endif /* QFQP_DEBUG */
