/*
 * Copyright (c) 2013 Carlo Augusto Grazia, Paolo Valente
 * Universita` di Modena e Reggio Emilia
 * All rights reserved 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifdef _KERNEL
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/module.h>
#include <net/if.h> /* IFNAMSIZ */
#include <netinet/in.h>
#include <netinet/ip_var.h> /* ipfw_rule_ref */
#include <netinet/ip_fw.h>  /* flow_id */
#include <netinet/ip_dummynet.h>
#include <netinet/ipfw/dn_heap.h>
#include <netinet/ipfw/ip_dn_private.h>
#include <netinet/ipfw/dn_sched.h>
#else
#include <dn_test.h>
#endif

#define DN_SINGLE   5 // XXX Where?


/* the scheduler algorithm */
struct dn_alg *p = NULL;


struct cfg_s {
    int ac;
    char * const *av;

    const char *name;

    /* running counters */
    uint32_t    _enqueue;
    uint32_t    drop;
    uint32_t    pending;
    uint32_t    dequeue;

    /* generator parameters */
    int flows;  /* number of flows */
    int flowsets;   /* number of flowsets */
    int wsum;   /* sum of weights of all flows */
    int max_flow_w; /* max flow weight */
    const char *fs_config; /* flowset config */

    struct mbuf *tosend; /* packet to send -- also flag to enqueue */

    struct mbuf *freelist;

    struct mbuf *head, *tail;   /* a simple tailq */

    int print_stat_level;

    /* scheduler hooks */
    int (*enq)(struct dn_sch_inst *, struct dn_queue *,
        struct mbuf *);
    struct mbuf * (*deq)(struct dn_sch_inst *);
    /* size of the three fields including sched-specific areas */
    int schk_len;
    int q_len; /* size of a queue including sched fields */
    int si_len; /* size of a sch_inst including sched fields */
    char *q;    /* array of flow queues */
        /* use a char* because size is variable */
    struct dn_fsk *fs;  /* array of flowsets */
    struct dn_sch_inst *si;
    struct dn_schk *sched;

    int single_queue; /*
                       * equal to 1 if the scheduler uses just one
                       * queue to store the packets of all the flows
                       */
} *sched;

/* global max queueing delay */
double global_queue = 0;

/* link speed in Kb/sec */
int bw_sched = 54000;

/* global Bit Worst-case Fair Index (B-WFI) */
double global_bwfi = 0;

/* global Time Worst-case Fair Index (T-WFI) */
double global_twfi = 0;

/*
 * FI2Q and Q2FI converts from flow_id to dn_queue and back.
 * We cannot easily use pointer arithmetic because it is variable size.
 */
#define FI2Q(c, i)  ((struct dn_queue *)((c)->q + (c)->q_len * (i)))
#define Q2FI(c, q)  (((char *)(q) - (c)->q)/(c)->q_len)

static int
single_drop(struct cfg_s *c, struct mbuf *m)
{
    m->m_nextpkt = c->freelist;
    c->freelist = m;
    return 0;
}

/* interpret a number in human form */
static long
getnum(const char *s, char **next, const char *key)
{
    char *end = NULL;
    long l;

    if (next)   /* default */
        *next = NULL;
    if (s && *s) {
        DX(3, "token is <%s> %s", s, key ? key : "-");
        l = strtol(s, &end, 0);
    } else {
        DX(3, "empty string");
        l = -1;
    }
    if (l < 0) {
        DX(2, "invalid %s for %s", s ? s : "NULL", (key ? key : "") );
        return 0;   /* invalid */
    }
    if (!end || !*end)
        return l;
    if (*end == 'n')
        l = -l; /* multiply by n */
    else if (*end == 'K')
        l = l*1000;
    else if (*end == 'M')
        l = l*1000000;
    else if (*end == 'k')
        l = l*1024;
    else if (*end == 'm')
        l = l*1024*1024;
    else if (*end == 'w')
        ;
    else {/* not recognized */
        D("suffix %s for %s, next %p", end, key, next);
        end--;
    }
    end++;
    DX(3, "suffix now %s for %s, next %p", end, key, next);
    if (next && *end) {
        DX(3, "setting next to %s for %s", end, key);
        *next = end;
    }
    return l;
}

/*
 * flowsets are a comma-separated list of
 * weight:maxlen:flows
 * indicating how many flows are hooked to that fs.
 * Both weight and range can be min-max-steps.
 * In a first pass we just count the number of flowsets and flows,
 * in a second pass we complete the setup.
 */

static void
parse_flowsets(struct cfg_s *c, const char *fs, int pass)
{
    char *s, *cur, *next;
    int n_flows = 0, n_fs = 0, wsum = 0;
    int i, j;
    struct dn_fs *prev = NULL;

    DX(3, "--- pass %d flows %d flowsets %d", pass, c->flows, c->flowsets);
    if (pass == 0)
        c->fs_config = fs;
    s = c->fs_config ? strdup(c->fs_config) : NULL;
    if (s == NULL) {
        if (pass == 0)
            D("no fsconfig");
        return;
    }
    for (next = s; (cur = strsep(&next, ","));) {
        char *p = NULL;
        int w, w_h, w_steps, wi;
        int len, len_h, l_steps, li;
        int flows;

        w = getnum(strsep(&cur, ":"), &p, "weight");
        if (w <= 0)
            w = 1;
        w_h = p ? getnum(p+1, &p, "weight_max") : w;
        w_steps = p ? getnum(p+1, &p, "w_steps") : (w_h == w ?1:2);
        len = getnum(strsep(&cur, ":"), &p, "len");
        if (len <= 0)
            len = 1000;
        len_h = p ? getnum(p+1, &p, "len_max") : len;
        l_steps = p ? getnum(p+1, &p, "l_steps") : (len_h == len ? 1 : 2);
        flows = getnum(strsep(&cur, ":"), NULL, "flows");
        if (flows == 0)
            flows = 1;
        DX(4, "weight %d..%d (%d) len %d..%d (%d) flows %d",
            w, w_h, w_steps, len, len_h, l_steps, flows);
        if (w == 0 || w_h < w || len == 0 || len_h < len ||
                flows == 0) {
            DX(4,"wrong parameters %s", fs);
            return;
        }
        n_flows += flows * w_steps * l_steps;
        for (i = 0; i < w_steps; i++) {
            wi = w + ((w_h - w)* i)/(w_steps == 1 ? 1 : (w_steps-1));
            for (j = 0; j < l_steps; j++, n_fs++) {
                struct dn_fs *fs = &c->fs[n_fs].fs; 
                int x;

                li = len + ((len_h - len)* j)/(l_steps == 1 ? 1 : (l_steps-1));
                x = (wi*2048)/li;
                DX(3, "----- fs %4d weight %4d lmax %4d X %4d flows %d",
                    n_fs, wi, li, x, flows);
                if (pass == 0)
                    continue;
                if (c->fs == NULL || c->flowsets <= n_fs) {
                    D("error in number of flowsets");
                    return;
                }
                wsum += wi * flows;
                fs->par[0] = wi;
                fs->par[1] = li > sizeof(struct mbuf) ?
			li : sizeof(struct mbuf);
                fs->index = n_fs;
                fs->n_flows = flows;
                fs->cur = fs->first_flow = prev==NULL ? 0 : prev->next_flow;
                fs->next_flow = fs->first_flow + fs->n_flows;
                fs->y = x * flows;
                fs->base_y = (prev == NULL) ? 0 : prev->next_y;
                fs->next_y = fs->base_y + fs->y;
                prev = fs;
            }
        }
    }
    c->flows = n_flows;
    c->flowsets = n_fs;
    c->wsum = wsum;
    if (pass == 0)
        return;

    /* now link all flows to their parent flowsets */
    DX(1,"%d flows on %d flowsets", c->flows, c->flowsets);
    for (i=0; i < c->flowsets; i++) {
        struct dn_fs *fs = &c->fs[i].fs;
        DX(1, "fs %3d w %5d l %4d flow %5d .. %5d y %6d .. %6d",
            i, fs->par[0], fs->par[1],
            fs->first_flow, fs->next_flow,
            fs->base_y, fs->next_y);
        for (j = fs->first_flow; j < fs->next_flow; j++) {
            struct dn_queue *q = FI2Q(c, j);
            q->fs = &c->fs[i];
        }
    }
}


/* enqueue returns NON-zero when a packet is dropped */
static int
enqueue(struct cfg_s *c, void *_m)
{
    int failure = 0;
    struct mbuf *m = _m;
    struct dn_queue *q = FI2Q(c, m->flow_id);

    if (c->enq) {
        failure = c->enq(c->si, q, m);
        if (! failure && c->single_queue)
            goto inc_length;
        else
            return failure;
    }
    
    if (c->head == NULL)
        c->head = m;
    else
        c->tail->m_nextpkt = m;
    c->tail = m;
inc_length:
    /*
     * Make sure that each flow is accounted for the number of
     * packets of that flow that are pending, even in case of
     * single_queue schedulers (which do not update the following
     * field)
     */
    q->ni.length++;
    return 0; /* default - success */
}

/* dequeue returns NON-NULL when a packet is available */
static void *
dequeue(struct cfg_s *c)
{
    struct mbuf *m;
    struct dn_queue *q;
    if (c->deq) {
        m = c->deq(c->si);
        if (m != NULL && c->single_queue)
            goto dec_length;
        else
            return m;
    }
    if ((m = c->head)) {
        m = c->head;
        c->head = m->m_nextpkt;
        m->m_nextpkt = NULL;
    }
dec_length:
    q = FI2Q(c, m->flow_id);
    /*
     * Make sure that each flow is accounted (only) for the number
     * of packets of that flow that are pending, even in case of
     * single_queue schedulers (which do not update the following
     * field)
     */
    q->ni.length--;
    return m;
}

/* Create packet */
static void *
create_pkt(int flow_id)
{
    struct mbuf *m;
    struct dn_queue *q;

    q = FI2Q(sched, flow_id);

    /* construct a packet */
    if (sched->freelist) {
        m = sched->tosend = sched->freelist;
        sched->freelist = sched->freelist->m_nextpkt;
    } else {
        m = sched->tosend = calloc(1, sizeof(struct mbuf));
    }
    if (m == NULL)
        return 0;

    m->Q_bklg = q->ni.length + 1; /* update cur. backlog, for T-WFI */
    m->enq_time = sched->dequeue;
    m->cfg = sched;
    m->m_nextpkt = NULL;
    m->m_pkthdr.len = sizeof(struct mbuf);
    m->flow_id = flow_id;
    m->father_flow_id = flow_id;

    return m;
}

static int 
single_enqueue(struct dn_sch_inst *_si, int flow_id)
{
    struct dn_queue *q;
    struct mbuf *m;

    sched->pending++;

    m = create_pkt(flow_id);

    q = FI2Q(sched, flow_id);
    /* Start new B-WFI update cycle if queue becomes backlogged */
    if(q->ni.length == 0){
        q->ni.allpkt_serv_bklgd = sched->dequeue;
        q->ni.pkt_serv_bklgd = q->ni.pkts;
    }

    /* update max len stats */
    if(q->ni.length + 1 > q->ni.max_length)
        q->ni.max_length = q->ni.length + 1;

    ND(" pending %d flow %5d len %4d",
        sched->pending, m->flow_id, q->ni.length);

    sched->_enqueue++;
    return enqueue(sched,m);
}

/* single_dequeue returns -1 when dequeue() fails */
static int
single_dequeue(struct dn_sch_inst *_si)
{
    int flow_id;
    struct mbuf *m;
    struct dn_queue *q;
    int queueing_delay;
    double wfi;
    double twfi;
    
    /* 
     * extract the macsal packet dequeue and recover 
     * the qos one associated to return 
     */
    if ((m = dequeue(sched))) {
        sched->dequeue++;
        flow_id = m->father_flow_id;

        /* STATS queueing delay */
        queueing_delay = sched->dequeue - m->enq_time;
        q = FI2Q(sched, flow_id);
        /* max time */
        if (queueing_delay > q->ni.q_time_max)
		q->ni.q_time_max = queueing_delay;
        /* min time */
        if(queueing_delay < q->ni.q_time_min || q->ni.q_time_min == 0)
            q->ni.q_time_min = queueing_delay;
        /* avg time */
        q->ni.q_avg_time += queueing_delay;
        /* time variance (at the end divided for all the packets dequeued)*/
        q->ni.q_time_square += (queueing_delay*queueing_delay);

        q->ni.pkts++;

        /* Update stas for B-WFI */
        if(q->ni.length == 0){ /* 
                                * queue no more backlogged, stop counters (t2) 
                                */
            q->ni.allpkt_serv_bklgd = sched->dequeue - q->ni.allpkt_serv_bklgd;
            q->ni.pkt_serv_bklgd = q->ni.pkts - q->ni.pkt_serv_bklgd;
            wfi = ((double)q->fs->fs.par[0]/sched->wsum) *
		   q->ni.allpkt_serv_bklgd - q->ni.pkt_serv_bklgd;
            if(wfi > global_bwfi)
                global_bwfi = wfi;
        }

        /* Update stats for T-WFI */
        twfi = queueing_delay -
		m->Q_bklg / ( (double)q->fs->fs.par[0] / sched->wsum);

        if (twfi > global_twfi && q->fs->fs.par[0] == sched->max_flow_w)
		global_twfi = twfi;

        /* END STATS time */
        ND(" deq, pending %d flow %5d",
            sched->pending, flow_id);
        sched->pending--;
        single_drop(sched,m); 
        sched->drop--;  
    } else {
        return -1;
    }

    return flow_id;
}


static int
single_new_sched(struct dn_sch_inst *_si)
{
    int i;
    int ac = _si->sched->fs->fs.n_args;
    char *const *av = _si->sched->fs->fs.args;

    moduledata_t *scheduler = NULL;

    /* Here we define the structure and the parameters for the MAC-SAL */
    sched = calloc(1, sizeof(struct cfg_s));
    sched->si_len = sizeof(struct dn_sch_inst);
    sched->q_len = sizeof(struct dn_queue);

    extern moduledata_t *_g_dn_fifo;        
    extern moduledata_t *_g_dn_wf2qp;
    extern moduledata_t *_g_dn_rr;
    extern moduledata_t *_g_dn_qfq;
    extern moduledata_t *_g_dn_qfqp;
    extern moduledata_t *_g_dn_kps;

    sched->flows = 1;
    sched->max_flow_w = 0;
    sched->flowsets = 1;
    sched->name = "null";
    ac--; av++;
    while (ac > 1) {
        if (!strcmp(*av, "-alg")) {
            if (!strcmp(av[1], "rr"))
                scheduler = _g_dn_rr;
            else if (!strcmp(av[1], "wf2qp"))
                scheduler = _g_dn_wf2qp;
            else if (!strcmp(av[1], "fifo"))
                scheduler = _g_dn_fifo;
            else if (!strcmp(av[1], "qfq"))
                scheduler = _g_dn_qfq;
            else if (!strcmp(av[1], "qfq+"))
                scheduler = _g_dn_qfqp;
            else if (!strcmp(av[1], "kps"))
                scheduler = _g_dn_kps;
            else
                scheduler = NULL;
            DX(3, "using scheduler %s", scheduler->name);
        } else if (!strcmp(*av, "-stats")) {
            sched->print_stat_level = getnum(av[1], NULL, av[0]);
            DX(3, "setting print stats to %d", sched->print_stat_level);
        } else if (!strcmp(*av, "-flows")) {
            sched->flows = getnum(av[1], NULL, av[0]);
            DX(3, "setting flows to %d", sched->flows);
        } else if (!strcmp(*av, "-flowsets")) {
            parse_flowsets(sched, av[1], 0);
            DX(3, "setting flowsets to %d", sched->flowsets);
        } else {
            DX(0,"option %s not recognised for macsal, ignore", *av);
        }
        ac -= 2; av += 2;
    } /* set the eventually wrong MAC-SAL parameter */
    if (sched->flows <= 0)
        sched->flows = 1;
    if (sched->flowsets <= 0)
        sched->flowsets = 1;

    /* Define the algorithm */
    if(scheduler)
        p = scheduler->p;
    else{
        printf("\nError, no -alg algo found\n");
        exit(1);
    }

    if (scheduler == NULL || scheduler == _g_dn_fifo)
        sched->single_queue = 1;

    if (scheduler) {
        sched->enq = p->enqueue;
        sched->deq = p->dequeue;
        sched->si_len += p->si_datalen;
        sched->q_len += p->q_datalen;
        sched->schk_len += p->schk_datalen;
    }

    /* allocate queues, flowsets and one sched for mac-sal */
    sched->q = calloc(sched->flows, sched->q_len);
    sched->fs = calloc(sched->flowsets, sizeof(struct dn_fsk));
    sched->si = calloc(1, sched->si_len);
    sched->sched = calloc(sched->flows, sched->schk_len);
    if (sched->q == NULL || sched->fs == NULL) {
        D("error allocating memory for flows");
        exit(1);
    }
    sched->si->sched = sched->sched;
    if (p) {
        if (p->config)
            p->config(sched->sched);
        if (p->new_sched)
            p->new_sched(sched->si);
    }

    /* parse_flowsets links queues to their flowsets */
    parse_flowsets(sched, av[1], 1);
    /* complete the work calling new_fsk for sched */
    for (i = 0; i < sched->flowsets; i++) {
        if (sched->fs[i].fs.par[1] == 0)
            sched->fs[i].fs.par[1] = sizeof(struct mbuf); /* default pkt len */
        sched->fs[i].sched = sched->sched;
        if (p && p->new_fsk)
            p->new_fsk(&sched->fs[i]);
    }


    /* 
     * initialize the lists for the generator, and put
     * all flows in the list for backlog = 0, for macsal
     */
    for (i = 0; i < sched->flows; i++) {
        struct dn_queue *q = FI2Q(sched, i);
        if (q->fs == NULL)
            q->fs = &sched->fs[0]; /* XXX */
        q->_si = sched->si;
        if (p && p->new_queue)
            p->new_queue(q);
    }


    for(i = 0; i < sched->flows; i++){
        struct dn_queue *q = FI2Q(sched, i);

        if(q->fs->fs.par[0] > sched->max_flow_w)
            sched->max_flow_w = q->fs->fs.par[0];
    }

    return 0;
}

/* update and print statistics at the end of the run */
static void
single_finalprint_stats(struct dn_sch_inst *_si)
{
    struct dn_queue *q;
    int i;

    double wfi, global_twfi_ms;
    /* 
     * Update stats for B-WFI to handle the case where t2 
     * falls on the end of experiment 
     */
    for(i = 0; i < sched->flows; i++){
            q = FI2Q(sched, i);
            if(q->ni.length != 0){ /*
				    * Pretend that the backlog is finished,
				    * to update the quantities used to
				    * compute B-WFI
				    */
		    q->ni.allpkt_serv_bklgd = sched->dequeue - 
			    q->ni.allpkt_serv_bklgd;
		    q->ni.pkt_serv_bklgd = q->ni.pkts - q->ni.pkt_serv_bklgd;
            }
    }

    /* calculate B-WFI for each flow */
    for(i = 0; i < sched->flows; i++){
            q = FI2Q(sched, i);
            wfi =((double)q->fs->fs.par[0]/(double)sched->wsum)*
		    (double)q->ni.allpkt_serv_bklgd -
		    (double)q->ni.pkt_serv_bklgd;
            if(wfi > global_bwfi)
		    global_bwfi = wfi;
    }

    /* update max queueing delay */
    for(i = 0; i < sched->flows; i++){
            q = FI2Q(sched, i);
            if( q->fs->fs.par[0] == sched->max_flow_w ){
		    if( ((double)q->ni.q_time_max * 8.0 * sizeof(struct mbuf))/
			bw_sched > global_queueing_delay )
			    global_queueing_delay =
				    ((double)q->ni.q_time_max * 8.0 * 
				     sizeof(struct mbuf))/bw_sched;
            }
    }

    /* turn into a time [ms] */
    global_twfi_ms = global_twfi * 8.0 * sizeof(struct mbuf) / bw_sched;

    if(sched->print_stat_level > 1){
        printf("%5.3f     %5.3f     %5.3f     %5.3f     %5.3f    ", 
               global_queueing_delay, global_bwfi, global_bwfi,
	       global_twfi_ms, global_twfi_ms);

	return;
    } else if (sched->print_stat_level == 0) {
	    D("\tqueuing delay: %5.3fL/R, B-WFI: %5.3fL, "
	      "T-WFI: %5.3fL/R <-> %5.3fms", 
	      global_queueing_delay, global_bwfi,
	      global_twfi, global_twfi_ms);

	return;
    }

    for(i = 0; i < sched->flows; i++){
    	double avg = 0;
    	double sigma = 0;

        q = FI2Q(sched, i);
        if(q->ni.pkts > 0) {
            avg = (double)q->ni.q_avg_time / q->ni.pkts;
            sigma = abs(q->ni.q_time_square/ (double)q->ni.pkts - avg*avg);
        }
        D("flow %d, pkts %d, min_del %d, max_del %d, avg_del %.2f, var %.2f," 
               "max_len %d", i, q->ni.pkts, q->ni.q_time_min, q->ni.q_time_max,
               avg, sigma, q->ni.max_length);
    }
}


/*
 * SINGLE scheduler descriptor
 * contains the type of the scheduler, the name, the size of the
 * structures and function pointers.
 */
static struct dn_alg single_desc = {
    .type = DN_SINGLE,
    .name = "SINGLE",
    .flags = DN_MULTIQUEUE,

    .si_datalen = sizeof(struct dn_queue),
    .q_datalen = sizeof(struct dn_queue),

    .enqueue_wrapper = single_enqueue,
    .dequeue_wrapper = single_dequeue,

    .new_sched = single_new_sched,
    .finalprint_stats = single_finalprint_stats,
};


DECLARE_DNSCHED_MODULE(dn_single, &single_desc);
